// Copyright 1999
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// MathUtil.h

///////////////////////////////////////////////////////////////////////////////

#ifndef MATHUTIL_H_
#define MATHUTIL_H_

#include "CHeaders.h"


// Useful numeric limits

const double			maximum_double			= numeric_limits<double>::			max();
const double			minimum_double			= - maximum_double;

const float				maximum_float			= numeric_limits<float>::			max();
const float				minimum_float			= - maximum_float;

const int				maximum_int				= numeric_limits<int>::				max();
const int				minimum_int				= numeric_limits<int>::				min();

const long				maximum_long			= numeric_limits<long>::			max();
const long				minimum_long			= numeric_limits<long>::			min();

const short				maximum_short			= numeric_limits<short>::			max();
const short				minimum_short			= numeric_limits<short>::			min();

const unsigned int		maximum_unsigned_int	= numeric_limits<unsigned int>::	max();

const unsigned long		maximum_unsigned_long	= numeric_limits<unsigned long>::	max();

const unsigned short	maximum_unsigned_short	= numeric_limits<unsigned short>::	max();


// Useful numeric test functions

inline bool ValidDouble(double x) { return (minimum_double <= x && x <= maximum_double); }

inline bool ValidFloat (float  x) { return (minimum_float  <= x && x <= maximum_float ); }


// Utility: SafeRoundToInt returns x safely rounded to an integer

inline int SafeRoundToInt(double d) {
	const double upper = maximum_int + 0.5;
	
	if (d >= upper)
		return maximum_int;
	
	const double lower = minimum_int - 0.5;
	
	if (d <= lower)
		return minimum_int;
	
	// use built-in conversion which truncates
	// use 0.5 offset to obtain round
	
	int   n = (d >= 0) ? (d + 0.5) : (d - 0.5);
	
	return n;
}


// Utility: SafeRoundToLong returns x safely rounded to a long integer

inline long SafeRoundToLong(double d) {
	const double upper = maximum_long + 0.5;
	
	if (d >= upper)
		return maximum_long;
	
	const double lower = minimum_long - 0.5;
	
	if (d <= lower)
		return minimum_long;
	
	// use built-in conversion which truncates
	// use 0.5 offset to obtain round
	
	long  n = (d >= 0) ? (d + 0.5) : (d - 0.5);
	
	return n;
}


// Utility: SafeRoundToShort returns d safely rounded to a short integer

inline short SafeRoundToShort(double d) {
	const double upper = maximum_short + 0.5;
	
	if (d >= upper)
		return maximum_short;
	
	const double lower = minimum_short - 0.5;
	
	if (d <= lower)
		return minimum_short;
	
	// use built-in conversion which truncates
	// use 0.5 offset to obtain round
	
	short n = (d >= 0) ? (d + 0.5) : (d - 0.5);
	
	return n;
}


// Utility: SafeRoundToUnsignedLong returns d safely rounded to an unsigned long integer

inline unsigned long SafeRoundToUnsignedLong(double d) {
	const double upper = maximum_unsigned_long + 0.5;
	
	if (d >= upper)
		return maximum_unsigned_long;
	
	const double lower = - 0.5;
	
	if (d <= lower)
		return 0;
	
	// use built-in conversion which truncates
	// use 0.5 offset to obtain round
	
	unsigned long n = d + 0.5;
	
	return n;
}


// For backward compatibility, define SafeRound:

inline short SafeRound(double d) { return SafeRoundToShort(d); }


// Utility: Ratio returns the double precision ratio of 2 long integers

inline double Ratio (long x, long y) {
	double xx = x;
	double yy = y;
	
	return xx/yy;
}


// Utility: Expand interval from a to b to include x

inline void ExpandInterval(double& a, double& b, double x) {
	// put a and b into the right order
	if (b < a) {
		double c = b;
		b = a;
		a = c;
	}
	
	// expand limits
	if (x < a)
		a = x;
	else
	if (x > b)
		b = x;
}


// Radians to Degrees Conversions

const double pi = 3.14159265358979324;

const double radians_to_degrees = 180 / pi;

const double degrees_to_radians = pi / 180;


// Omit: Summer 97
// const double onedegree = pi/180;


// Trigonometric Functions in Degrees

inline double sindeg(double x) {
	return sin(x * degrees_to_radians);
}

inline double cosdeg(double x) {
	return cos(x * degrees_to_radians);
}

inline double tandeg(double x) {
	return tan(x * degrees_to_radians);
}

inline double asindeg(double x) {
	return asin(x) * radians_to_degrees;
}

inline double acosdeg(double x) {
	return acos(x) * radians_to_degrees;
}

inline double atandeg(double x) {
	return atan(x) * radians_to_degrees;
}

// notice that the two parameter form of the
// arc tangent takes y before x

inline double atan2deg(double y, double x) {
	return atan2(y, x) * radians_to_degrees;
}


// Natural Logarithm Synonym

inline double ln(double x) { return log(x); }


// Utilities for Scaling


// compute the linear scale factor that transforms
// range a,b into range ia,ib

inline double ComputeFactor(double a, double b, short ia, short ib) {
	if ((a == b) || (ia == ib))
		return 1;
	else
		return double(ib - ia) / (b - a);
}


// compute the linear scale offset that transforms
// range a,b into range ia,ib using the given factor
// guarantee that (a + b)/2 maps to double(ia + ib)/2

inline double ComputeOffset(double a, double b, short ia, short ib, double factor) {
	return (double(ia + ib) - factor * (a + b)) / 2;
}


// equalize two numbers to the smaller magnitude
// do nothing however if either number is zero

template <class Number>
inline void EqualizeSize(Number& a, Number& b) {
	if ((a == 0) || (b == 0))
		return;
	
	// compute absolute values
	
	Number A = (a > 0) ? a : (-a);
	Number B = (b > 0) ? b : (-b);
	
	// equalize to the smaller magnitude
	
	if (A < B)
		b = (b > 0) ? A : (-A);
	else
		a = (a > 0) ? B : (-B);
}

#endif // MATHUTIL_H_